from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict

from .utils import from_json


@dataclass  # type: ignore
class BaseMessage(ABC):
    prefix: str
    content: str

    def __init__(self, content: str, prefix: str) -> None:
        self.content = content
        self.prefix = prefix

    def to_json(self) -> Dict[str, str]:
        return {"role": self.prefix, "content": self.content}

    def copy(self) -> "BaseMessage":
        return self.from_json(self.to_json())

    @staticmethod
    @abstractmethod
    def from_json(json: Dict[str, str]) -> "BaseMessage":
        return from_json(json)  # type: ignore

    def __str__(self) -> str:
        return f"{self.prefix}: {self.content}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.prefix}, {self.content})"



from typing import Dict


def from_json(json: Dict[str, str]):  # type: ignore
    """Return a BaseMessage from a JSON dict."""
    from . import AssistantMessage, SystemMessage, UserMessage

    if json["role"] == "user":
        return UserMessage.from_json(json)
    elif json["role"] == "system":
        return SystemMessage.from_json(json)
    elif json["role"] == "assistant":
        return AssistantMessage.from_json(json)
    else:
        raise ValueError(f"Invalid role: {json['role']}")



from .assistant import AssistantMessage
from .base import BaseMessage
from .system import SystemMessage
from .user import UserMessage

__all__ = ["AssistantMessage", "SystemMessage", "UserMessage", "BaseMessage"]



from abc import ABC, abstractmethod
from typing import Any, Dict, Generator, Union

from tinylang.memory import ConversationMemory
from tinylang.memory.base import BaseMemory


class BaseLLM(ABC):
    """
    Abstract Base Class for Large Language Models.
    """

    def __init__(self, memory: BaseMemory | None = None) -> None:
        self.memory = memory or ConversationMemory()

    @abstractmethod
    def load_model(self, model_path: str) -> bool:
        """
        Loads the model from the specified path into memory.

        :param model_path: Path to the model.
        """
        raise NotImplementedError

    @abstractmethod
    def chat(
        self, prompt: str, stream: bool = False, raw_response: bool = False
    ) -> Union[str, Generator[Dict[str, Any], None, None]]:
        """
        Simulates a chat interaction with the model.

        :param prompt: The user prompt.
        :param stream: Whether to stream the response or not.
        :param raw_response: Whether to return the raw response or not.
        :return: Generated output text.
        """
        raise NotImplementedError



from typing import Dict

from tinylang.messages.base import BaseMessage


class UserMessage(BaseMessage):
    def __init__(self, content: str, prefix: str = "user") -> None:
        super().__init__(content, prefix)

    @staticmethod
    def from_json(json: Dict[str, str]) -> "UserMessage":
        return UserMessage(content=json["content"])



from collections import deque
from dataclasses import dataclass
from typing import Dict, List

from ..messages import AssistantMessage, SystemMessage, UserMessage
from ..messages.base import BaseMessage
from .base import BaseMemory


@dataclass
class ConversationMemory(BaseMemory):
    last_k: int | None

    def __init__(
        self,
        messages: List[BaseMessage] | None = None,
        last_k: int | None = None,
    ) -> None:
        self._last_k_must_be_positive_and_even(last_k)
        super().__init__(messages)

        self.last_k = last_k or 1
        self._last_k_messages = self.last_k * 2  # keep assistant and user messages

        # use a deque to keep the last k messages
        self.messages = deque(self.messages, maxlen=self._last_k_messages)

    @staticmethod
    def _last_k_must_be_positive_and_even(last_k: int | None) -> int | None:
        if last_k is None:
            return last_k

        if last_k < 1:
            raise ValueError("last_k must be positive")

        # must be even
        if last_k % 2 == 1:
            raise ValueError("last_k must be even")

        return last_k

    def add_message(self, message: BaseMessage) -> None:
        self.messages.append(message)

    def add_user_message(self, message: str) -> None:
        self.add_message(UserMessage(message))

    def add_system_message(self, message: str) -> None:
        self.add_message(SystemMessage(message))

    def add_assistant_message(self, message: str) -> None:
        self.add_message(AssistantMessage(message))

    def to_json(self) -> List[Dict[str, str]]:
        return [message.to_json() for message in self.messages]

    def format_messages(self) -> List[Dict[str, str]]:
        temp = [message.to_json() for message in self.messages]
        return temp[-self._last_k_messages :]



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

__version__ = "0.4.4"

setup(
    name="tinylang",
    version=__version__,
    install_requires=requirements,
    packages=find_packages(),
    description="A tiny language interpreter",
    long_description=long_description,
    long_description_content_type="text/markdown",
)



from typing import Dict

from .base import BaseMessage


class AssistantMessage(BaseMessage):
    def __init__(self, content: str, prefix: str = "assistant") -> None:
        super().__init__(content, prefix)

    @staticmethod
    def from_json(json: Dict[str, str]) -> "AssistantMessage":
        return AssistantMessage(content=json["content"])



"""Base class for memory."""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, List, MutableSequence

from ..messages.base import BaseMessage


@dataclass  # type: ignore
class BaseMemory(ABC):
    def __init__(self, messages: MutableSequence[BaseMessage] | None = None) -> None:
        self.messages = messages or []

    @abstractmethod
    def add_message(self, message: BaseMessage) -> None:
        raise NotImplementedError

    @abstractmethod
    def add_user_message(self, message: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def add_system_message(self, message: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def add_assistant_message(self, message: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def to_json(self) -> List[Dict[str, str]]:
        raise NotImplementedError

    def format_messages(self) -> List[Dict[str, str]]:
        return [message.to_json() for message in self.messages]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.messages})"

    def __str__(self) -> str:
        return str(self.messages)



from typing import Dict

from .base import BaseMessage


class SystemMessage(BaseMessage):
    def __init__(self, content: str, prefix: str = "system") -> None:
        super().__init__(content, prefix)

    @staticmethod
    def from_json(json: Dict[str, str]) -> "SystemMessage":
        return SystemMessage(content=json["content"])



from typing import Any, Dict, Generator, Union

import openai
from openai.openai_object import OpenAIObject

from tinylang.llms.base import BaseLLM
from tinylang.memory.base import BaseMemory


class OpenAI(BaseLLM):
    def __init__(
        self,
        openai_api_key: str,
        openai_organization: str,
        model: str,
        memory: BaseMemory | None = None,
        **kwargs: Dict,
    ) -> None:
        super().__init__(memory)
        self.openai_api_key = openai_api_key
        self.openai_organization = openai_organization
        self.model = model
        self.kwargs = kwargs

        openai.api_key = self.openai_api_key
        openai.organization = self.openai_organization

    def load_model(self, model_path: str) -> bool:
        return True  # nothing to do here since using APImake

    def _stream_response(
        self, api_response: OpenAIObject, raw_response: bool
    ) -> Generator[Dict[str, Any], None, None]:
        aggregated_content = ""
        for chunk in api_response:
            content = chunk["choices"][0]["delta"].get("content")

            # if there's no content
            if not content:
                continue

            aggregated_content += content

            yield chunk if raw_response else content

        self.memory.add_assistant_message(aggregated_content)

    def chat(
        self, prompt: str, stream: bool = False, raw_response: bool = False
    ) -> Union[str, Generator[Dict[str, Any], None, None]]:
        self.memory.add_user_message(prompt)

        api_response: OpenAIObject = openai.ChatCompletion.create(
            model=self.model,
            messages=self.memory.format_messages(),
            stream=stream,
            **self.kwargs,
        )

        if stream:
            return self._stream_response(api_response, raw_response)
        else:
            chat_response: str = api_response["choices"][0]["message"]["content"]
            self.memory.add_assistant_message(chat_response)
            return chat_response
